@@ -1,38 +1,42 @@
-#!/usr/bin/env python
+#!/usr/bin/env python2
 
 import sys
 reload(sys)
 sys.setdefaultencoding('utf-8')
 
-try:
-    sys.path.append('/usr/lib/cinnamon-settings/modules')
-    sys.path.append('/usr/lib/cinnamon-settings/bin')
-    import os
-    import glob
-    import gettext
-    from gi.repository import Gio, Gtk, GObject, GdkPixbuf, GLib, Pango, Gdk, cairo
-    import SettingsWidgets
-    import capi
-    import time
-    import grp
-    import pwd
-# Standard setting pages... this can be expanded to include applet dirs maybe?
-    mod_files = glob.glob('/usr/lib/cinnamon-settings/modules/*.py')
-    mod_files.sort()
-    if len(mod_files) is 0:
-        raise Exception("No settings modules found!!")
-    for i in range(len(mod_files)):
-        mod_files[i] = mod_files[i].split('/')[5]
-        mod_files[i] = mod_files[i].split('.')[0]
-        if mod_files[i][0:3] != "cs_":
-            raise Exception("Settings modules must have a prefix of 'cs_' !!")
-    modules = map(__import__, mod_files)
-except Exception, detail:
-    print detail
-    sys.exit(1)
+sys.path.append('/usr/lib/cinnamon-settings/modules')
+sys.path.append('/usr/lib/cinnamon-settings/bin')
+import os
+import glob
+import gettext
+from gi.repository import Gio, Gtk, GObject, GdkPixbuf, GLib, Pango, Gdk, cairo
+import SettingsWidgets
+import capi
+import time
+import traceback
+import locale
+import urllib2
+import proxygsettings
+from functools import cmp_to_key
 
 # i18n
-gettext.install("cinnamon", "/usr/share/cinnamon/locale")
+gettext.install("cinnamon", "/usr/share/locale")
+
+# Standard setting pages... this can be expanded to include applet dirs maybe?
+mod_files = glob.glob('/usr/lib/cinnamon-settings/modules/*.py')
+mod_files.sort()
+if len(mod_files) is 0:
+    print "No settings modules found!!"
+    sys.exit(1)
+
+mod_files = [x.split('/')[5].split('.')[0] for x in mod_files]
+
+for mod_file in mod_files:
+    if mod_file[0:3] != "cs_":
+        raise Exception("Settings modules must have a prefix of 'cs_' !!")
+
+modules = map(__import__, mod_files)
+
 # i18n for menu item
 menuName = _("System Settings")
 menuComment = _("Control Center")
@@ -46,6 +50,8 @@ MAX_LABEL_WIDTH = 25
 MIN_PIX_WIDTH = 100
 MAX_PIX_WIDTH = 160
 
+MOUSE_BACK_BUTTON = 8
+
 CATEGORIES = [
 #        Display name                         ID              Show it? Always False to start              Icon
     {"label": _("Appearance"),            "id": "appear",      "show": False,                       "icon": "cs-cat-appearance"},
@@ -58,8 +64,8 @@ CONTROL_CENTER_MODULES = [
 #         Label                              Module ID                Icon                         Category      Keywords for filter
     [_("Networking"),                       "network",            "cs-network",                 "hardware",      _("network, wireless, wifi, ethernet, broadband, internet")],
     [_("Display"),                          "display",            "cs-display",                 "hardware",      _("display, screen, monitor, layout, resolution, dual, lcd")],
-    [_("Bluetooth"),                        "bluetooth",          "cs-bluetooth",               "hardware",      _("bluetooth, dongle, transfer, mobile")], 
-    [_("Accessibility"),                 "universal-access",   "cs-universal-access",           "prefs",         _("magnifier, talk, access, zoom, keys, contrast")],
+    [_("Bluetooth"),                        "bluetooth",          "cs-bluetooth",               "hardware",      _("bluetooth, dongle, transfer, mobile")],
+    [_("Accessibility"),                    "universal-access",   "cs-universal-access",        "prefs",         _("magnifier, talk, access, zoom, keys, contrast")],
     [_("Sound"),                            "sound",              "cs-sound",                   "hardware",      _("sound, speakers, headphones, test")],
     [_("Color"),                            "color",              "cs-color",                   "hardware",      _("color, profile, display, printer, output")],
     [_("Graphics Tablet"),                  "wacom",              "cs-tablet",                  "hardware",      _("wacom, digitize, tablet, graphics, calibrate, stylus")]
@@ -67,14 +73,14 @@ CONTROL_CENTER_MODULES = [
 
 STANDALONE_MODULES = [
 #         Label                          Executable                          Icon                Category        Keywords for filter
-    [_("Printers"),                      "system-config-printer",        "cs-printer",         "hardware",       _("printers, laser, inkjet")],    
+    [_("Printers"),                      "system-config-printer",        "cs-printer",         "hardware",       _("printers, laser, inkjet")],
     [_("Firewall"),                      "gufw",                         "cs-firewall",        "admin",          _("firewall, block, filter, programs")],
     [_("Languages"),                     "mintlocale",                   "cs-language",        "prefs",          _("language, install, foreign")],
-    [_("Login Screen"),                  "gksu /usr/sbin/mdmsetup",      "cs-login",           "admin",          _("login, mdm, gdm, manager, user, password, startup, switch")],
-    [_("Startup Programs"),              "cinnamon-session-properties",  "cs-startup-programs","prefs",          _("startup, programs, boot, init, session")],
-    [_("Device Drivers"),                "mintdrivers",                  "cs-drivers",         "admin",          _("video, driver, wifi, card, hardware, proprietary, nvidia, radeon, nouveau, fglrx")],
+    [_("Login Window"),                  "gksu /usr/sbin/mdmsetup",      "cs-login",           "admin",          _("login, mdm, gdm, manager, user, password, startup, switch")],
+    [_("Driver Manager"),                "mintdrivers",                  "cs-drivers",         "admin",          _("video, driver, wifi, card, hardware, proprietary, nvidia, radeon, nouveau, fglrx")],
     [_("Software Sources"),              "mintsources",                  "cs-sources",         "admin",          _("ppa, repository, package, source, download")],
-    [_("Users and Groups"),              "cinnamon-settings-users",      "cs-user-accounts",   "admin",          _("user, users, account, accounts, group, groups, password")]
+    [_("Users and Groups"),              "cinnamon-settings-users",      "cs-user-accounts",   "admin",          _("user, users, account, accounts, group, groups, password")],
+    [_("Bluetooth"),                     "blueberry",                    "cs-bluetooth",       "hardware",       _("bluetooth, dongle, transfer, mobile")]
 ]
 
 def print_timing(func):
@@ -100,31 +106,65 @@ class MainWindow:
             filtered_path = side_view.get_model().convert_path_to_child_path(selected_items[0])
             if filtered_path is not None:
                 self.go_to_sidepage(cat, filtered_path)
+    
+    def _on_sidepage_hide_stack(self):
+        self.stack_switcher.set_opacity(0)
+        
+    def _on_sidepage_show_stack(self):
+        self.stack_switcher.set_opacity(1)
 
-    def go_to_sidepage(self, cat, path):        
+    def go_to_sidepage(self, cat, path):
         iterator = self.store[cat].get_iter(path)
         sidePage = self.store[cat].get_value(iterator,2)
         if not sidePage.is_standalone:
-            self.side_view_sw.hide()
-            self.search_entry.hide()
             self.window.set_title(sidePage.name)
             sidePage.build()
-            self.content_box_sw.show()
-            self.button_back.show()
+            if sidePage.stack:
+                self.stack_switcher.set_stack(sidePage.stack)
+                if sidePage.stack.get_visible():
+                    self.stack_switcher.set_opacity(1)
+                else:
+                    self.stack_switcher.set_opacity(0)
+                if hasattr(sidePage, "connect_proxy"):
+                    sidePage.connect_proxy("hide_stack", self._on_sidepage_hide_stack)
+                    sidePage.connect_proxy("show_stack", self._on_sidepage_show_stack)
+            else:
+                self.stack_switcher.set_opacity(0)
+            self.main_stack.set_visible_child_name("content_box_page")
+            self.header_stack.set_visible_child_name("content_box")
             self.current_sidepage = sidePage
+            width = 0
+            for widget in self.top_bar:
+                m, n = widget.get_preferred_width()
+                width += n
+            self.top_bar.set_size_request(width + 20, -1)
             self.maybe_resize(sidePage)
         else:
             sidePage.build()
 
     def maybe_resize(self, sidePage):
         m, n = self.content_box.get_preferred_size()
+
+        # Resize horizontally if the module is wider than the window
         use_width = WIN_WIDTH
         if n.width > WIN_WIDTH:
             use_width = n.width
+
+        # Resize vertically depending on the height requested by the module
+        use_height = WIN_HEIGHT
         if not sidePage.size:
-            self.window.resize(use_width, n.height + self.bar_heights + WIN_H_PADDING)
-        elif sidePage.size > -1:
-            self.window.resize(use_width, sidePage.size + self.bar_heights + WIN_H_PADDING)
+            # No height requested, resize vertically if the module is taller than the window
+            if n.height > WIN_HEIGHT:
+                use_height = n.height + self.bar_heights + WIN_H_PADDING
+            #self.window.resize(use_width, n.height + self.bar_heights + WIN_H_PADDING)
+        elif sidePage.size > 0:
+            # Height hardcoded by the module
+            use_height = sidePage.size + self.bar_heights + WIN_H_PADDING
+        elif sidePage.size == -1:
+            # Module requested the window to fit it (i.e. shrink the window if necessary)
+            use_height = n.height + self.bar_heights + WIN_H_PADDING
+
+        self.window.resize(use_width, use_height)
 
     def deselect(self, cat):
         for key in self.side_view.keys():
@@ -134,58 +174,86 @@ class MainWindow:
     ''' Create the UI '''
     @print_timing
     def __init__(self):
-
         self.builder = Gtk.Builder()
         self.builder.add_from_file("/usr/lib/cinnamon-settings/cinnamon-settings.ui")
         self.window = self.builder.get_object("main_window")
         self.top_bar = self.builder.get_object("top_bar")
         self.side_view = {}
+        self.main_stack = self.builder.get_object("main_stack")
+        self.main_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
+        self.main_stack.set_transition_duration(150)
+        self.header_stack = self.builder.get_object("header_stack")
+        self.header_stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
+        self.header_stack.set_transition_duration(150)
         self.side_view_container = self.builder.get_object("category_box")
         self.side_view_sw = self.builder.get_object("side_view_sw")
         self.side_view_sw.show_all()
         self.content_box = self.builder.get_object("content_box")
         self.content_box_sw = self.builder.get_object("content_box_sw")
+        self.content_box_sw.show_all()
         self.button_back = self.builder.get_object("button_back")
-        self.button_back.set_label(_("All Settings"))
-        self.button_back.hide()
+        self.button_back.set_tooltip_text(_("Back to all settings"))
+        button_image = self.builder.get_object("image1")
+        button_image.props.icon_size = Gtk.IconSize.MENU
+
+        self.stack_switcher = self.builder.get_object("stack_switcher")
+        # Set stack to random thing and make opacity 0 so that the heading bar
+        # does not resize when switching between pages
+        self.stack_switcher.set_stack(self.main_stack)
+
+        m, n = self.button_back.get_preferred_width()
+        self.stack_switcher.set_margin_right(n)
 
         self.search_entry = self.builder.get_object("search_box")
         self.search_entry.connect("changed", self.onSearchTextChanged)
         self.search_entry.connect("icon-press", self.onClearSearchBox)
+
         self.window.connect("destroy", self.quit)
         self.window.connect("key-press-event", self.on_keypress)
+        self.window.connect("button-press-event", self.on_buttonpress)
         self.window.show()
 
         self.builder.connect_signals(self)
         self.window.set_has_resize_grip(False)
+        self.unsortedSidePages = []
         self.sidePages = []
         self.settings = Gio.Settings.new("org.cinnamon")
-        self.current_cat_widget = None            
+        self.current_cat_widget = None
 
         self.current_sidepage = None
         self.c_manager = capi.CManager()
         self.content_box.c_manager = self.c_manager
         self.bar_heights = 0
 
-        for i in range(len(modules)):
+        for module in modules:
             try:
-                mod = modules[i].Module(self.content_box)
+                mod = module.Module(self.content_box)
                 if self.loadCheck(mod) and self.setParentRefs(mod):
-                    self.sidePages.append((mod.sidePage, mod.name, mod.category))
+                    self.unsortedSidePages.append((mod.sidePage, mod.name, mod.category))
             except:
-                print "Failed to load module %s" % modules[i]
-                import traceback
+                print "Failed to load module %s" % module
                 traceback.print_exc()
 
         for item in CONTROL_CENTER_MODULES:
             ccmodule = SettingsWidgets.CCModule(item[0], item[1], item[2], item[3], item[4], self.content_box)
             if ccmodule.process(self.c_manager):
-                self.sidePages.append((ccmodule.sidePage, ccmodule.name, ccmodule.category))
+                self.unsortedSidePages.append((ccmodule.sidePage, ccmodule.name, ccmodule.category))
 
         for item in STANDALONE_MODULES:
             samodule = SettingsWidgets.SAModule(item[0], item[1], item[2], item[3], item[4], self.content_box)
             if samodule.process():
-                self.sidePages.append((samodule.sidePage, samodule.name, samodule.category))
+                self.unsortedSidePages.append((samodule.sidePage, samodule.name, samodule.category))
+
+        # sort the modules alphabetically according to the current locale
+        sidePageNamesToSort = map(lambda m: m[0].name, self.unsortedSidePages)
+        sortedSidePageNames = sorted(sidePageNamesToSort, key=cmp_to_key(locale.strcoll))
+        for sidePageName in sortedSidePageNames:
+            nextSidePage = None
+            for trySidePage in self.unsortedSidePages:
+                if(trySidePage[0].name == sidePageName):
+                    nextSidePage = trySidePage
+
+            self.sidePages.append(nextSidePage);
 
 
         # create the backing stores for the side nav-view.
@@ -229,7 +297,6 @@ class MainWindow:
 
         # set up larger components.
         self.window.set_title(_("System Settings"))
-        self.window.connect("destroy", self.quit)
         self.button_back.connect('clicked', self.back_to_icon_view)
 
         self.calculate_bar_heights()
@@ -246,11 +313,21 @@ class MainWindow:
             self.search_entry.grab_focus()
 
     def on_keypress(self, widget, event):
-        if event.keyval == Gdk.KEY_BackSpace and (type(self.window.get_focus()) not in
-                                (Gtk.TreeView, Gtk.Entry, Gtk.SpinButton, Gtk.TextView)):
+        grab = False
+        device = Gtk.get_current_event_device()
+        if device.get_source() == Gdk.InputSource.KEYBOARD:
+            grab = Gdk.Display.get_default().device_is_grabbed(device)
+        if not grab and event.keyval == Gdk.KEY_BackSpace and (type(self.window.get_focus()) not in
+                    (Gtk.TreeView, Gtk.Entry, Gtk.SpinButton, Gtk.TextView)):
             self.back_to_icon_view(None)
             return True
-        return False    
+        return False
+
+    def on_buttonpress(self, widget, event):
+        if event.button == MOUSE_BACK_BUTTON:
+            self.back_to_icon_view(None)
+            return True
+        return False
 
     def calculate_bar_heights(self):
         h = 0
@@ -267,7 +344,7 @@ class MainWindow:
 
     def filter_visible_function(self, model, iter, user_data = None):
         sidePage = model.get_value(iter, 2)
-        text = self.search_entry.get_text().lower()       
+        text = self.search_entry.get_text().lower()
         if sidePage.name.lower().find(text) > -1 or \
            sidePage.keywords.lower().find(text) > -1:
             return True
@@ -343,6 +420,7 @@ class MainWindow:
         area.pack_start(text_renderer, True, True, False)
         area.add_attribute(pixbuf_renderer, "icon-name", 1)
         pixbuf_renderer.set_property("stock-size", Gtk.IconSize.DIALOG)
+        pixbuf_renderer.set_property("follow-state", True)
 
         area.add_attribute(text_renderer, "text", 0)
 
@@ -473,7 +551,7 @@ class MainWindow:
 
     def setParentRefs (self, mod):
         try:
-            mod._setParentRef(self.window, self.builder)
+            mod._setParentRef(self.window)
         except AttributeError:
             pass
         return True
@@ -487,7 +565,6 @@ class MainWindow:
     def back_to_icon_view(self, widget):
         self.window.set_title(_("System Settings"))
         self.window.resize(WIN_WIDTH, WIN_HEIGHT)
-        self.content_box_sw.hide()
         children = self.content_box.get_children()
         for child in children:
             child.hide()
@@ -495,18 +572,25 @@ class MainWindow:
                 c_widgets = child.get_children()
                 for c_widget in c_widgets:
                     c_widget.hide()
-        self.button_back.hide()
-        self.side_view_sw.show()
-        self.search_entry.show()
+        self.main_stack.set_visible_child_name("side_view_page")
+        self.header_stack.set_visible_child_name("side_view")
         self.search_entry.grab_focus()
-        self.current_sidepage = None   
-    
-    def quit(self, *args):
-        Gtk.main_quit()
+        self.current_sidepage = None
 
+    def quit(self, *args):
+        self.window.destroy()
+        Gtk.main_quit()
 
 if __name__ == "__main__":
     import signal
-    signal.signal(signal.SIGINT, MainWindow().quit)
-    Gtk.main()
 
+    ps = proxygsettings.get_proxy_settings()
+    if ps:
+        proxy = urllib2.ProxyHandler(ps)
+    else:
+        proxy = urllib2.ProxyHandler()
+    urllib2.install_opener(urllib2.build_opener(proxy))
+
+    window = MainWindow()
+    signal.signal(signal.SIGINT, window.quit)
+    Gtk.main()
